from tornado import ioloop

import kiwipy

__all__ = ['new_event_loop', 'set_event_loop', 'get_event_loop']

get_event_loop = ioloop.IOLoop.current
new_event_loop = ioloop.IOLoop


def set_event_loop(loop):
    if loop is None:
        ioloop.IOLoop.clear_instance()
    else:
        loop.make_current()



import tornado.concurrent
from functools import partial

__all__ = ['Future', 'gather', 'chain', 'copy_future', 'InvalidStateError', 'CancelledError']


class InvalidStateError(BaseException):
    """The operation is not allowed in this state."""
    pass


class CancelledError(Exception):
    """The operation as cancelled ."""
    pass


class Future(tornado.concurrent.Future):
    _cancelled = False

    def set_result(self, result):
        if self.done():
            raise InvalidStateError('Future already done')
        super(Future, self).set_result(result)

    def cancel(self):
        if self.done():
            return False

        self._cancelled = True
        # Get the callbacks scheduled
        self._set_done()
        return True

    def cancelled(self):
        return self._cancelled

    def result(self):
        if self.cancelled():
            raise CancelledError
        if not self.done():
            raise InvalidStateError("Not done yet so no result")

        return super(Future, self).result(timeout=0.)


def copy_future(source, target):
    """ Copy the status of future a to b unless b is already done in
    which case return

    :param source: The source future
    :type source: :class:`Future`
    :param target: The target future
    :type target: :class:`Future`
    """

    if target.done():
        return

    if source.cancelled():
        target.cancel()
    else:
        if source.exc_info() is not None:
            target.set_exc_info(source.exc_info())
        else:
            target.set_result(source.result())


def chain(a, b):
    """Chain two futures together so that when one completes, so does the other.

    The result (success or failure) of ``a`` will be copied to ``b``, unless
    ``b`` has already been completed or cancelled by the time ``a`` finishes.
    """

    a.add_done_callback(lambda first: copy_future(first, b))


def gather(*args):
    if not args:
        future = Future()
        future.set_result([])
        return future
    return _GatheringFuture(*args)


class _GatheringFuture(Future):
    def __init__(self, *args):
        super(_GatheringFuture, self).__init__()
        self._children = list(args)
        self._nchildren = len(self._children)
        self._nfinished = 0
        self._result = [None] * self._nchildren

        for i, future in enumerate(self._children):
            future.add_done_callback(partial(self._completed, i))

    def cancel(self):
        for child in self._children:
            child.cancel()

    def _completed(self, i, future):
        if self.cancelled():
            return

        if future.cancelled():
            self.cancel()
        else:
            if future.exception() is not None:
                self._result[i] = future.exception()
            else:
                self._result[i] = future.result()

            # Check if we're all done
            self._nfinished += 1
            if self._nfinished == self._nchildren:
                self.set_result(self._result)



from future.utils import with_metaclass
import abc
import kiwipy


class CommunicatorTester(with_metaclass(abc.ABCMeta)):
    WAIT_TIMEOUT = 0.5

    def setUp(self):
        super(CommunicatorTester, self).setUp()
        self.communicator = self.create_communicator()

    def tearDown(self):
        self.destroy_communicator(self.communicator)

    @abc.abstractmethod
    def create_communicator(self):
        """
        :return: A constructed communicator
        :rtype: :class:`kiwi.Communicator`
        """
        pass

    def destroy_communicator(self, communicator):
        pass

    def test_rpc_send_receive(self):
        MESSAGE = "sup yo'"
        RESPONSE = "nuthin bra"

        messages = []

        def on_receive(msg):
            messages.append(msg)
            return RESPONSE

        self.communicator.add_rpc_subscriber(on_receive, 'rpc')
        response = self.communicator.rpc_send_and_wait('rpc', MESSAGE)

        self.assertEqual(messages[0], MESSAGE)
        self.assertEqual(response, RESPONSE)

    def test_task_send(self):
        TASK = 'The meaning?'
        RESULT = 42

        tasks = []

        def on_task(task):
            tasks.append(task)
            return RESULT

        self.communicator.add_task_subscriber(on_task)
        result = self.communicator.task_send_and_wait(TASK)

        self.assertEqual(tasks[0], TASK)
        self.assertEqual(result, RESULT)

    def test_future_task(self):
        TASK = 'The meaning?'
        RESULT = 42
        result_future = kiwipy.Future()

        tasks = []

        def on_task(task):
            tasks.append(task)
            return result_future

        self.communicator.add_task_subscriber(on_task)
        task_future = self.communicator.task_send(TASK)

        result_future.set_result(42)

        result = self.communicator.await(task_future)

        self.assertEqual(tasks[0], TASK)
        self.assertEqual(result, RESULT)

    def test_task_exception(self):
        TASK = 'The meaning?'

        tasks = []

        def on_task(task):
            tasks.append(task)
            raise RuntimeError("I cannea do it Captain!")

        self.communicator.add_task_subscriber(on_task)
        with self.assertRaises(kiwipy.RemoteException):
            self.communicator.task_send_and_wait(TASK)

        self.assertEqual(tasks[0], TASK)

    def test_broadcast_send(self):
        SUBJECT = 'yo momma'
        BODY = 'so fat'
        SENDER_ID = 'me'
        FULL_MSG = {'body': BODY, 'subject': SUBJECT, 'sender': SENDER_ID, 'correlation_id': None}

        message1 = kiwipy.Future()
        message2 = kiwipy.Future()

        def on_broadcast_1(*args, **msg):
            message1.set_result(msg)

        def on_broadcast_2(*args, **msg):
            message2.set_result(msg)

        self.communicator.add_broadcast_subscriber(on_broadcast_1)
        self.communicator.add_broadcast_subscriber(on_broadcast_2)

        self.communicator.broadcast_send(**FULL_MSG)
        # Wait fot the send and receive
        self.communicator.await(kiwipy.gather(message1, message2), timeout=self.WAIT_TIMEOUT)

        self.assertDictEqual(message1.result(), FULL_MSG)
        self.assertDictEqual(message2.result(), FULL_MSG)

    def test_broadcast_filter_subject(self):
        subjects = []
        EXPECTED_SUBJECTS = ['purchase.car', 'purchase.piano']

        done = kiwipy.Future()

        def on_broadcast_1(body, sender=None, subject=None, correlation_id=None):
            subjects.append(subject)
            if len(subjects) == len(EXPECTED_SUBJECTS):
                done.set_result(True)

        self.communicator.add_broadcast_subscriber(
            kiwipy.BroadcastFilter(on_broadcast_1, subject="purchase.*"))

        for subject in ['purchase.car', 'purchase.piano', 'sell.guitar', 'sell.house']:
            self.communicator.broadcast_send(None, subject=subject)

        self.communicator.await(done, timeout=self.WAIT_TIMEOUT)

        self.assertEqual(len(subjects), 2)
        self.assertListEqual(EXPECTED_SUBJECTS, subjects)

    def test_broadcast_filter_sender(self):
        EXPECTED_SENDERS = ['bob.jones', 'alice.jones']
        senders = []

        done = kiwipy.Future()

        def on_broadcast_1(body, sender=None, subject=None, correlation_id=None):
            senders.append(sender)
            if len(senders) == len(EXPECTED_SENDERS):
                done.set_result(True)

        self.communicator.add_broadcast_subscriber(
            kiwipy.BroadcastFilter(on_broadcast_1, sender="*.jones"))

        for sender in ['bob.jones', 'bob.smith', 'martin.uhrin', 'alice.jones']:
            self.communicator.broadcast_send(None, sender=sender)

        self.communicator.await(done, timeout=self.WAIT_TIMEOUT)

        self.assertEqual(2, len(senders))
        self.assertListEqual(EXPECTED_SENDERS, senders)

    def test_broadcast_filter_sender_and_subject(self):
        senders_and_subects = set()
        EXPECTED = {
            ('bob.jones', 'purchase.car'),
            ('bob.jones', 'purchase.piano'),
            ('alice.jones', 'purchase.car'),
            ('alice.jones', 'purchase.piano'),
        }

        done = kiwipy.Future()

        def on_broadcast_1(body, sender=None, subject=None, correlation_id=None):
            senders_and_subects.add((sender, subject))
            if len(senders_and_subects) == len(EXPECTED):
                done.set_result(True)

        filtered = kiwipy.BroadcastFilter(on_broadcast_1)
        filtered.add_sender_filter("*.jones")
        filtered.add_subject_filter("purchase.*")
        self.communicator.add_broadcast_subscriber(filtered)

        for sender in ['bob.jones', 'bob.smith', 'martin.uhrin', 'alice.jones']:
            for subject in ['purchase.car', 'purchase.piano', 'sell.guitar', 'sell.house']:
                self.communicator.broadcast_send(None, sender=sender, subject=subject)

        self.communicator.await(done, timeout=self.WAIT_TIMEOUT)

        self.assertEqual(4, len(senders_and_subects))
        self.assertSetEqual(EXPECTED, senders_and_subects)

    def test_add_remove_broadcast_subscriber(self):
        broadcast_received = kiwipy.Future()

        def broadcast_subscriber(body, sender=None, subject=None, correlation_id=None):
            broadcast_received.set_result(True)

        # Check we're getting messages
        self.communicator.add_broadcast_subscriber(broadcast_subscriber)
        self.communicator.broadcast_send(None)
        self.assertTrue(self.communicator.await(broadcast_received, timeout=self.WAIT_TIMEOUT))

        self.communicator.remove_broadcast_subscriber(broadcast_subscriber)
        # Check that we're unsubscribed
        broadcast_received = kiwipy.Future()
        with self.assertRaises(kiwipy.TimeoutError):
            self.communicator.await(broadcast_received, timeout=self.WAIT_TIMEOUT)

    def test_add_remove_rpc_subscriber(self):
        def rpc_subscriber(msg):
            return True

        # Check we're getting messages
        self.communicator.add_rpc_subscriber(rpc_subscriber, rpc_subscriber.__name__)
        result = self.communicator.rpc_send_and_wait(rpc_subscriber.__name__, None, timeout=self.WAIT_TIMEOUT)
        self.assertTrue(result)

        self.communicator.remove_rpc_subscriber(rpc_subscriber.__name__)
        # Check that we're unsubscribed
        with self.assertRaises((kiwipy.UnroutableError, kiwipy.TimeoutError)):
            self.communicator.rpc_send_and_wait(rpc_subscriber.__name__, None, timeout=self.WAIT_TIMEOUT)



from functools import partial
import kiwipy
import logging
import pika
import pika.exceptions
import tornado.gen
from tornado.gen import coroutine
import tornado.ioloop
import yaml

from . import defaults
from . import tasks
from . import messages
from . import pubsub
from . import utils

__all__ = ['RmqCommunicator']

_LOGGER = logging.getLogger(__name__)

# The exchange properties use by the publisher and subscriber.  These have to match
# which is why they're declare her
EXCHANGE_PROPERTIES = {
    'exchange_type': 'topic',
}


class RmqPublisher(messages.BasePublisherWithReplyQueue):
    """

    """
    DEFAULT_EXCHANGE_PARAMS = EXCHANGE_PROPERTIES

    def rpc_send(self, recipient_id, msg):
        message = messages.RpcMessage(recipient_id, body=msg)
        self.action_message(message)
        return message.future

    def broadcast_send(self, msg, sender=None, subject=None, correlation_id=None):
        message = messages.BroadcastMessage(
            body=msg,
            sender=sender,
            subject=subject,
            correlation_id=correlation_id)
        self.action_message(message)


class RmqSubscriber(pubsub.ConnectionListener):
    def __init__(self, connector,
                 exchange_name=defaults.TASK_EXCHANGE,
                 decoder=yaml.load,
                 encoder=yaml.dump):
        """
        Subscribes and listens for process control messages and acts on them
        by calling the corresponding methods of the process manager.

        :param connector: The RMQ connector
        :param exchange_name: The name of the exchange to use
        :param decoder:
        """
        super(RmqSubscriber, self).__init__()

        self._connector = connector
        self._exchange_name = exchange_name
        self._decode = decoder
        self._response_encode = encoder
        self._channel = None

        self._rpc_subscribers = {}
        self._broadcast_subscribers = []

        self._connector = connector
        self._active = False

    def add_rpc_subscriber(self, subscriber, identifier):
        self._rpc_subscribers[identifier] = subscriber

    def remove_rpc_subscriber(self, identifier):
        try:
            self._rpc_subscribers.pop(identifier)
        except KeyError:
            raise ValueError("Unknown subscriber '{}'".format(identifier))

    def add_broadcast_subscriber(self, subscriber):
        self._broadcast_subscribers.append(subscriber)

    def remove_broadcast_subscriber(self, subscriber):
        self._broadcast_subscribers.remove(subscriber)

    def on_connection_opened(self, connector, connection):
        self.connect()

    def channel(self):
        return self._channel

    @coroutine
    def connect(self):
        if self._channel:
            # Already connected
            return

        connector = self._connector
        channel = yield connector.open_channel()
        channel.add_on_close_callback(self._on_channel_closed)
        yield connector.exchange_declare(channel, exchange=self._exchange_name, **EXCHANGE_PROPERTIES)

        # RPC queue
        frame = yield connector.queue_declare(
            channel,
            exclusive=True,
            arguments={
                "x-message-ttl": defaults.MESSAGE_TTL,
                "x-expires": defaults.QUEUE_EXPIRES
            })
        rpc_queue = frame.method.queue
        result = yield connector.queue_bind(
            channel,
            queue=rpc_queue,
            exchange=self._exchange_name,
            routing_key='{}.*'.format(defaults.RPC_TOPIC))
        channel.basic_consume(self._on_rpc, queue=rpc_queue)

        # Broadcast queue
        frame = yield connector.queue_declare(
            channel,
            exclusive=True,
            arguments={
                "x-message-ttl": defaults.MESSAGE_TTL,
                "x-expires": defaults.QUEUE_EXPIRES
            })
        broadcast_queue = frame.method.queue
        yield connector.queue_bind(
            channel,
            queue=broadcast_queue,
            exchange=self._exchange_name,
            routing_key=defaults.BROADCAST_TOPIC)
        channel.basic_consume(self._on_broadcast, queue=broadcast_queue)

        self._channel = channel
        # Have we been called externally?  In which case activate
        if not self._active:
            self._connector.add_connection_listener(self)
            self._active = True

    @coroutine
    def disconnect(self):
        self._connector.remove_connection_listener(self)
        if self.channel() is not None:
            yield self._connector.close_channel(self.channel())
            self._channel = None

    # region RMQ methods

    def _on_channel_closed(self, ch, reply_code, reply_text):
        assert self._channel is ch
        self._channel = None

    # end region

    @coroutine
    def _on_rpc(self, ch, method, props, body):
        identifier = method.routing_key[len('{}.'.format(defaults.RPC_TOPIC)):]
        receiver = self._rpc_subscribers.get(identifier, None)
        if receiver is None:
            self._channel.basic_reject(method.delivery_tag)
        else:
            # Tell the sender that we've dealt with it
            self._channel.basic_ack(method.delivery_tag)

            msg = self._decode(body)

            try:
                if tornado.gen.is_coroutine_function(receiver):
                    result = yield receiver(msg)
                else:
                    result = receiver(msg)

                if isinstance(result, tornado.concurrent.Future):
                    response = utils.pending_response()
                    self._send_response(ch, props.reply_to, props.correlation_id, response)
                    try:
                        response = yield result
                        response = utils.result_response(response)
                    except kiwipy.CancelledError as e:
                        response = utils.cancelled_response()
                    except BaseException as e:
                        response = utils.exception_response(e)
                else:
                    response = utils.result_response(result)
            except BaseException as e:
                response = utils.exception_response(e)

            self._send_response(ch, props.reply_to, props.correlation_id, response)

    @coroutine
    def _on_broadcast(self, ch, method, props, body):
        msg = self._decode(body)
        for receiver in self._broadcast_subscribers:
            try:
                if tornado.gen.is_coroutine_function(receiver):
                    yield receiver(**msg)
                else:
                    receiver(**msg)
            except BaseException:
                import traceback
                _LOGGER.error("Exception in broadcast receiver!\n"
                              "msg: {}\n"
                              "traceback:\n{}".format(msg, traceback.format_exc()))

    def _send_response(self, ch, reply_to, correlation_id, response):
        ch.basic_publish(
            exchange='', routing_key=reply_to,
            properties=pika.BasicProperties(correlation_id=correlation_id),
            body=self._response_encode(response)
        )


class RmqCommunicator(kiwipy.Communicator):
    """
    A publisher and subscriber that implements the Communicator interface.
    """

    def __init__(self, connector,
                 exchange_name=defaults.MESSAGE_EXCHANGE,
                 task_exchange=defaults.TASK_EXCHANGE,
                 task_queue=defaults.TASK_QUEUE,
                 encoder=yaml.dump,
                 decoder=yaml.load,
                 task_prefetch_size=defaults.TASK_PREFETCH_SIZE,
                 task_prefetch_count=defaults.TASK_PREFETCH_COUNT,
                 testing_mode=False):
        """

        :param connector: The RMQ connector object
        :param exchange_name: The name of the RMQ message exchange to use
        :type exchange_name: str
        :param task_exchange: The name of the RMQ task exchange to use
        :type task_exchange: str
        :param task_queue: The name of the task queue to use
        :type task_queue: str
        :param encoder: The encoder to call for encoding a message
        :param decoder: The decoder to call for decoding a message
        :param testing_mode: Run in testing mode: all queues and exchanges
            will be temporary
        """
        super(RmqCommunicator, self).__init__()

        self._connector = connector
        self._loop = connector._loop
        self._connected = False

        self._message_subscriber = RmqSubscriber(
            connector,
            exchange_name,
            encoder=encoder,
            decoder=decoder
        )
        self._message_publisher = RmqPublisher(
            connector,
            exchange_name=exchange_name,
            encoder=encoder,
            decoder=decoder,
        )
        self._task_subscriber = tasks.RmqTaskSubscriber(
            connector,
            exchange_name=task_exchange,
            task_queue_name=task_queue,
            encoder=encoder,
            decoder=decoder,
            prefetch_size=task_prefetch_size,
            prefetch_count=task_prefetch_count,
            testing_mode=testing_mode,
        )
        self._task_publisher = tasks.RmqTaskPublisher(
            connector,
            exchange_name=task_exchange,
            task_queue_name=task_queue,
            encoder=encoder,
            decoder=decoder,
            testing_mode=testing_mode,
        )

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self):
        if self._connected:
            return

        self._loop.run_sync(self._message_subscriber.connect)
        self._loop.run_sync(self._task_subscriber.connect)
        self._loop.run_sync(self._message_publisher.connect)
        self._loop.run_sync(self._task_publisher.connect)

        self._connected = True

    def disconnect(self):
        self._loop.run_sync(self._message_publisher.disconnect)
        self._loop.run_sync(self._message_subscriber.disconnect)
        self._loop.run_sync(self._task_publisher.disconnect)
        self._loop.run_sync(self._task_subscriber.disconnect)
        self._loop.run_sync(self._connector.disconnect)

    def add_rpc_subscriber(self, subscriber, identifier):
        self._message_subscriber.add_rpc_subscriber(subscriber, identifier)

    def remove_rpc_subscriber(self, identifier):
        self._message_subscriber.remove_rpc_subscriber(identifier)

    def add_task_subscriber(self, subscriber):
        self._task_subscriber.add_task_subscriber(subscriber)

    def remove_task_subscriber(self, subscriber):
        self._task_subscriber.remove_task_subscriber(subscriber)

    def add_broadcast_subscriber(self, subscriber):
        self._message_subscriber.add_broadcast_subscriber(subscriber)

    def remove_broadcast_subscriber(self, subscriber):
        self._message_subscriber.remove_broadcast_subscriber(subscriber)

    def rpc_send(self, recipient_id, msg):
        """
        Initiate a remote procedure call on a recipient

        :param recipient_id: The recipient identifier
        :param msg: The body of the message
        :return: A future corresponding to the outcome of the call
        """
        try:
            return self._message_publisher.rpc_send(recipient_id, msg)
        except pika.exceptions.UnroutableError as e:
            raise kiwipy.UnroutableError(str(e))

    def broadcast_send(self, body, sender=None, subject=None, correlation_id=None):
        return self._message_publisher.broadcast_send(body, sender, subject, correlation_id)

    def task_send(self, msg):
        try:
            return self._task_publisher.task_send(msg)
        except pika.exceptions.UnroutableError as e:
            raise kiwipy.UnroutableError(str(e))
        except pika.exceptions.NackError as e:
            raise kiwipy.TaskRejected(str(e))

    def await(self, future=None, timeout=None):
        # Ensure we're connected
        if future is None:
            future = kiwipy.Future()
        self.connect()
        try:
            return self._loop.run_sync(lambda: future, timeout=timeout)
        except tornado.ioloop.TimeoutError as e:
            raise kiwipy.TimeoutError(str(e))



import logging
import uuid
from functools import partial
import kiwipy
import pika
from tornado.gen import coroutine
import yaml

from . import defaults
from . import messages
from . import pubsub
from . import utils

_LOGGER = logging.getLogger(__name__)

__all__ = ['RmqTaskSubscriber', 'RmqTaskPublisher']


class TaskMessage(messages.Message):
    def __init__(self, body, correlation_id=None):
        super(TaskMessage, self).__init__()
        self.correlation_id = correlation_id if correlation_id is not None else str(uuid.uuid4())
        self.body = body
        self._future = kiwipy.Future()

    @property
    def future(self):
        return self._future

    def send(self, publisher):
        if self.correlation_id is None:
            self.correlation_id = str(uuid.uuid4())
        publisher.publish_msg(
            self.body,
            routing_key=None,  # Set by the publisher
            correlation_id=self.correlation_id,
            mandatory=True, )

        publisher.await_response(self.correlation_id, self.on_response)
        return self.future

    def on_response(self, done_future):
        kiwipy.copy_future(done_future, self.future)


class RmqTaskSubscriber(messages.BaseConnectionWithExchange):
    """
    Listens for tasks coming in on the RMQ task queue
    """

    def __init__(self, connector,
                 task_queue_name=defaults.TASK_QUEUE,
                 testing_mode=False,
                 decoder=yaml.load,
                 encoder=yaml.dump,
                 exchange_name=defaults.MESSAGE_EXCHANGE,
                 exchange_params=None,
                 prefetch_size=defaults.TASK_PREFETCH_SIZE,
                 prefetch_count=defaults.TASK_PREFETCH_COUNT,
                 ):
        """
        :param connector: An RMQ connector
        :type connector: :class:`pubsub.RmqConnector`
        :param task_queue_name: The name of the queue to use
        :param decoder: A message decoder
        :param encoder: A response encoder
        """
        super(RmqTaskSubscriber, self).__init__(
            connector,
            exchange_name=exchange_name,
            exchange_params=exchange_params
        )

        self._task_queue = task_queue_name
        self._testing_mode = testing_mode
        self._decode = decoder
        self._encode = encoder
        self._prefetch_size = prefetch_size
        self._prefetch_count = prefetch_count

        self._subscribers = []
        self._pending_tasks = []

    def add_task_subscriber(self, subscriber):
        if not self._subscribers:
            self._consumer_tag = self.channel().basic_consume(self._on_task, self._task_queue)
        self._subscribers.append(subscriber)

    def remove_task_subscriber(self, subscriber):
        self._subscribers.remove(subscriber)
        if not self._subscribers:
            self.channel().basic_cancel(self._consumer_tag)

    def on_connection_closed(self, connector, reconnecting):
        for task in self._pending_tasks:
            try:
                task.cancel()
            except kiwipy.InvalidStateError:
                pass

    @coroutine
    def connect(self):
        if self.channel():
            # Already connected
            return

        yield super(RmqTaskSubscriber, self).connect()
        connector = self._connector
        self.channel().basic_qos(
            prefetch_count=self._prefetch_count,
            prefetch_size=self._prefetch_size)

        # Set up task queue
        task_queue = self._task_queue
        yield connector.queue_declare(
            self._channel,
            queue=task_queue,
            durable=not self._testing_mode,
            auto_delete=self._testing_mode,
            arguments={"x-message-ttl": defaults.TASK_MESSAGE_TTL}
        )
        # x-expires means how long does the queue stay alive after no clients
        # x-message-ttl means what is the default ttl for a message arriving in the queue
        yield connector.queue_bind(
            self._channel,
            queue=task_queue,
            exchange=self._exchange_name,
            routing_key=task_queue)

    @coroutine
    def _on_task(self, ch, method, props, body):
        handled = False
        for subscriber in self._subscribers:
            try:
                task = self._decode(body)
                result = subscriber(task)
                if isinstance(result, kiwipy.Future):
                    task_future = result
                    try:
                        self._pending_tasks.append(task_future)
                        result = yield task_future
                        response = utils.result_response(result)
                    except kiwipy.CancelledError:
                        response = utils.cancelled_response()
                    finally:
                        self._pending_tasks.remove(task_future)
                else:
                    response = utils.result_response(result)

                # Finished
                handled = True
                break
            except kiwipy.TaskRejected:
                pass
            except KeyboardInterrupt:
                raise
            except Exception as e:
                import traceback
                response = utils.exception_response('{}\n{}'.format(e, traceback.format_exc()))
                self._task_finished(method.delivery_tag, props.correlation_id, props.reply_to, response)
                raise

        if handled:
            self._task_finished(method.delivery_tag, props.correlation_id, props.reply_to, response)
        else:
            self._channel.basic_reject(delivery_tag=method.delivery_tag)

    def _task_finished(self, delivery_tag, correlation_id, reply_to, response):
        """
        Send an acknowledgement of the task being actioned and a response to the
        initiator.

        :param props: The message properties
        :param method: The message method
        :param response: The response to send to the initiator
        """
        self._channel.basic_ack(delivery_tag=delivery_tag)
        self._send_response(correlation_id, reply_to, response)

    def _send_response(self, correlation_id, reply_to, response):
        # Build full response
        response[utils.HOST_KEY] = utils.get_host_info()
        self.channel().basic_publish(
            exchange='',
            routing_key=reply_to,
            body=self._encode(response),
            properties=pika.BasicProperties(correlation_id=correlation_id))


class RmqTaskPublisher(messages.BasePublisherWithReplyQueue):
    """
    Publishes messages to the RMQ task queue and gets the response
    """

    def __init__(self, connector,
                 task_queue_name=defaults.TASK_QUEUE,
                 exchange_name=defaults.MESSAGE_EXCHANGE,
                 exchange_params=None,
                 encoder=yaml.dump,
                 decoder=yaml.load,
                 confirm_deliveries=True,
                 testing_mode=False):
        super(RmqTaskPublisher, self).__init__(
            connector,
            exchange_name=exchange_name,
            exchange_params=exchange_params,
            encoder=encoder,
            decoder=decoder,
            confirm_deliveries=confirm_deliveries,
            testing_mode=testing_mode)
        self._task_queue = task_queue_name

    def publish_msg(self, task, routing_key, correlation_id=None, mandatory=False, ttl=None):
        if routing_key is not None:
            _LOGGER.warn(
                "Routing key '{}' passed but is ignored for all tasks".format(routing_key))
        # pika (and AMQP) expects the ttl to be a string
        if ttl is not None and not isinstance(ttl, basestring):
            ttl = str(ttl)

        return self.do_publish(
            correlation_id,
            exchange=self.get_exchange_name(),
            routing_key=self._task_queue,
            properties=pika.BasicProperties(
                reply_to=self.get_reply_queue_name(),
                delivery_mode=2,  # Persistent
                correlation_id=correlation_id
            ),
            body=self._encode(task),
            mandatory=mandatory,
        )

    def task_send(self, msg):
        message = TaskMessage(msg)
        self.action_message(message)
        return message.future



import abc
from future.utils import with_metaclass
import sys

from . import futures

__all__ = ['Communicator', 'CommunicatorHelper',
           'RemoteException', 'DeliveryFailed', 'TaskRejected', 'UnroutableError',
           'TimeoutError'
           ]


class RemoteException(Exception):
    """ An exception occurred at the remote end of the call """
    pass


class DeliveryFailed(Exception):
    """ Failed to deliver a message """
    pass


class UnroutableError(DeliveryFailed):
    """ The messages was unroutable """
    pass


class TaskRejected(Exception):
    """ A task was rejected at the remote end """
    pass


class TimeoutError(Exception):
    """ Waiting for a future timed out """
    pass


class Communicator(with_metaclass(abc.ABCMeta)):
    """
    The interface for a communicator used to both send and receive various
    types of message.
    """

    @abc.abstractmethod
    def add_rpc_subscriber(self, subscriber, identifier):
        pass

    @abc.abstractmethod
    def remove_rpc_subscriber(self, identifier):
        """
        Remove an RPC subscriber given the identifier.  Raises a `ValueError` if there
        is no such subscriber.

        :param identifier: The RPC subscriber identifier
        """
        pass

    @abc.abstractmethod
    def add_task_subscriber(self, subscriber):
        pass

    @abc.abstractmethod
    def remove_task_subscriber(self, subscriber):
        pass

    @abc.abstractmethod
    def add_broadcast_subscriber(self, subscriber):
        pass

    @abc.abstractmethod
    def remove_broadcast_subscriber(self, subscriber):
        pass

    @abc.abstractmethod
    def task_send(self, msg):
        """
        Send a task messages, this will be queued and picked up by a
        worker at some point in the future.  The method returns a future
        representing the outcome of the task.

        :param msg: The task message
        :return: A future corresponding to the outcome of the task
        :rtype: :class:`kiwi.Future`
        """

    def task_send_and_wait(self, msg):
        future = self.task_send(msg)
        self.await(future)
        return future.result()

    @abc.abstractmethod
    def rpc_send(self, recipient_id, msg):
        """
        Initiate a remote procedure call on a recipient.  This method
        returns a future representing the outcome of the call.

        :param recipient_id: The recipient identifier
        :param msg: The body of the message
        :return: A future corresponding to the outcome of the call
        :rtype: :class:`kiwi.Future`
        """
        pass

    def rpc_send_and_wait(self, recipient_id, msg, timeout=None):
        future = self.rpc_send(recipient_id, msg)
        self.await(future, timeout=timeout)
        return future.result()

    @abc.abstractmethod
    def broadcast_send(self, body, sender=None, subject=None, correlation_id=None):
        pass

    @abc.abstractmethod
    def await(self, future=None, timeout=None):
        pass


class CommunicatorHelper(Communicator):
    def __init__(self):
        self._task_subscribers = []
        self._broadcast_subscribers = []
        self._rpc_subscribers = {}

    def add_rpc_subscriber(self, subscriber, identifier):
        self._rpc_subscribers[identifier] = subscriber

    def remove_rpc_subscriber(self, identifier):
        try:
            self._rpc_subscribers.pop(identifier)
        except KeyError:
            raise ValueError("Unknown subscriber '{}'".format(identifier))

    def add_task_subscriber(self, subscriber):
        """
        Register a task subscriber

        :param subscriber: The task callback function
        """
        self._task_subscribers.append(subscriber)

    def remove_task_subscriber(self, subscriber):
        """
        Deregister a task subscriber

        :param subscriber: The task callback function
        """
        # TODO: Put exception guard and raise out own exception
        self._task_subscribers.remove(subscriber)

    def add_broadcast_subscriber(self, broadcast_subscriber):
        self._broadcast_subscribers.append(broadcast_subscriber)

    def remove_broadcast_subscriber(self, broadcast_subscriber):
        self._broadcast_subscribers.remove(broadcast_subscriber)

    def fire_task(self, msg):
        future = futures.Future()
        handled = False

        for subscriber in self._task_subscribers:
            try:
                result = subscriber(msg)
                if isinstance(result, futures.Future):
                    future = result
                else:
                    future.set_result(result)

                handled = True
                break
            except TaskRejected:
                pass
            except Exception:
                future.set_exception(RemoteException(sys.exc_info()))
                handled = True

        if not handled:
            future.set_exception(TaskRejected("Rejected by all subscribers"))

        return future

    def fire_rpc(self, recipient_id, msg):
        try:
            subscriber = self._rpc_subscribers[recipient_id]
        except KeyError:
            raise UnroutableError("Unknown rpc recipient '{}'".format(recipient_id))
        else:
            future = futures.Future()
            try:
                result = subscriber(msg)
                if isinstance(result, futures.Future):
                    futures.chain(result, future)
                else:
                    future.set_result(result)
            except Exception:
                future.set_exception(RemoteException(sys.exc_info()))
            return future

    def fire_broadcast(self, body, sender=None, subject=None, correlation_id=None):
        for subscriber in self._broadcast_subscribers:
            subscriber(body=body, sender=sender, subject=subject, correlation_id=correlation_id)
        future = futures.Future()
        future.set_result(True)
        return future



import logging
import tempfile
import os

testfile = os.path.join(tempfile.gettempdir(), 'kiwipy_unittest.log')
try:
    os.remove(testfile)
except OSError:
    pass
print("Logging test to '{}'".format(testfile))
FORMAT = "[%(filename)s:%(lineno)s - %(funcName)s()] %(message)s"
logging.basicConfig(filename=testfile, level=logging.INFO, format=FORMAT)



import time

from . import communications
from . import futures

__all__ = ['LocalCommunicator']


class LocalCommunicator(communications.CommunicatorHelper):
    def task_send(self, msg):
        return self.fire_task(msg)

    def rpc_send(self, recipient_id, msg):
        return self.fire_rpc(recipient_id, msg)

    def broadcast_send(self, body, sender=None, subject=None, correlation_id=None):
        return self.fire_broadcast(body, sender=sender, subject=subject, correlation_id=correlation_id)

    def await(self, future=None, timeout=None):
        if not future.done():
            time.sleep(timeout)
        try:
            return future.result()
        except futures.InvalidStateError:
            raise communications.TimeoutError("Timed out waiting for result")



from contextlib import contextmanager
from enum import Enum
import logging
import pika
import pika.exceptions
import kiwipy

from . import loops

__all__ = ['RmqConnector', 'ConnectionListener']

LOGGER = logging.getLogger(__name__)


class ConnectionListener(object):
    def on_connection_opened(self, connector, connection):
        pass

    def on_connection_closed(self, connector, reconnecting):
        pass


class ConnectorState(Enum):
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2
    WAITING_TO_RECONNECT = 3


class RmqConnector(object):
    """
    A basic RMQ client that opens a connection and one channel.
    If an auto reconnect timeout is given it will try to keep the connection
    open by reopening if it is closed.
    """
    _connection = None

    def __init__(self,
                 amqp_url,
                 auto_reconnect_timeout=None,
                 loop=None):
        self._connection_params = pika.URLParameters(amqp_url)
        self._reconnect_timeout = auto_reconnect_timeout
        self._loop = loop if loop is not None else loops.new_event_loop()

        self._event_helper = kiwipy.EventHelper(ConnectionListener)
        self._running_future = None
        self._stopping = False

        self._connecting_future = kiwipy.Future()
        self._disconnecting_future = None

        self._state = ConnectorState.DISCONNECTED

    @property
    def is_connected(self):
        return self._connection is not None and self._connection.is_open

    def get_blocking_connection(self):
        return pika.BlockingConnection(self._connection_params)

    @contextmanager
    def blocking_channel(self, confirm_delivery=True):
        conn = self.get_blocking_connection()
        channel = conn.channel()
        if confirm_delivery:
            channel.confirm_delivery()
        try:
            yield channel
        finally:
            channel.close()
            conn.close()

    def get_connection_params(self):
        return self._connection_params

    def connect(self):
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika. If you want the reconnection to work, make
        sure you set stop_ioloop_on_close to False, which is not the default
        behavior of this adapter.
        """
        if self._state is ConnectorState.CONNECTED:
            return self._connecting_future

        LOGGER.info('Connecting to %s', self.get_connection_params())
        if self._state is ConnectorState.DISCONNECTED:
            self._state = ConnectorState.CONNECTING
            self._connection = pika.TornadoConnection(
                self._connection_params,
                on_open_callback=self._on_connection_open,
                on_close_callback=self._on_connection_closed,
                custom_ioloop=self._loop)

        return self._connecting_future

    def disconnect(self):
        """This method closes the connection to RabbitMQ."""
        if self._state is not ConnectorState.DISCONNECTED:
            LOGGER.info('Closing connection')
            self._disconnecting_future = kiwipy.Future()
            self._connection.close()
            self._connection = None
            self._connecting_future = None
            self._state = ConnectorState.DISCONNECTED

        return self._disconnecting_future

    def open_channel(self):
        connection_future = self.connect()
        future = kiwipy.Future()
        connection_future.add_done_callback(
            lambda _: self._open_channel(connection_future.result(), future))
        return future

    def _open_channel(self, connection, future):
        """This method will open a new channel with RabbitMQ by issuing the
        Channel.Open RPC command.
        """
        LOGGER.info('Creating a new channel')
        connection.channel(on_open_callback=future.set_result)
        return future

    def close_channel(self, channel):
        LOGGER.info("Closing channel {}".format(channel))
        future = kiwipy.Future()
        channel.add_on_close_callback(lambda *args: future.set_result(args))
        channel.close()
        return future

    def exchange_declare(self, channel, nowait=False, **kwargs):
        params = dict(kwargs)
        params['nowait'] = nowait
        if nowait:
            callback = None
            future = None
        else:
            future = kiwipy.Future()
            callback = future.set_result
        params['callback'] = callback

        LOGGER.info('Declaring exchange {}'.format(params))
        channel.exchange_declare(**params)
        return future

    def queue_declare(self, channel, nowait=False, **kwargs):
        params = dict(kwargs)
        params['nowait'] = nowait
        if nowait:
            callback = None
            future = None
        else:
            future = kiwipy.Future()
            callback = future.set_result
        params['callback'] = callback

        LOGGER.info('Declaring queue {}'.format(params))
        channel.queue_declare(**params)
        return future

    def queue_bind(self, channel, nowait=False, **kwargs):
        params = dict(kwargs)
        params['nowait'] = nowait
        if nowait:
            callback = None
            future = None
        else:
            future = kiwipy.Future()
            callback = future.set_result
        params['callback'] = callback

        LOGGER.info('Binding queue {}'.format(params))
        channel.queue_bind(**params)
        return future

    def connection(self):
        return self._connection

    def add_connection_listener(self, listener):
        self._event_helper.add_listener(listener)

    def remove_connection_listener(self, listener):
        self._event_helper.remove_listener(listener)

    def _on_connection_open(self, connection):
        """Called when the RMQ connection has been opened

        :type connection: pika.BaseConnection
        """
        LOGGER.info('Connection opened')
        self._connection = connection
        self._connecting_future.set_result(connection)
        self._state = ConnectorState.CONNECTED

        self._event_helper.fire_event(ConnectionListener.on_connection_opened, self, connection)

    def _on_connection_closed(self, connection, reply_code, reply_text):
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.

        :param pika.connection.Connection connection: The closed connection obj
        :param int reply_code: The server provided reply_code if given
        :param str reply_text: The server provided reply_text if given

        """
        self._connection = None

        LOGGER.info("The connection was closed: ({}) {}".format(reply_code, reply_text))

        if self._state is not ConnectorState.DISCONNECTED and self._reconnect_timeout is not None:
            self._connecting_future = kiwipy.Future()
            self._state = ConnectorState.WAITING_TO_RECONNECT
            LOGGER.warning(
                "Connection closed, reopening in {} seconds: ({}) {}".format(
                    self._reconnect_timeout, reply_code, reply_text
                ))
            self._connection.add_timeout(self._reconnect_timeout, self._reconnect)
        else:
            self._state = ConnectorState.DISCONNECTED
            self._connecting_future = None
            if self._disconnecting_future:
                self._disconnecting_future.set_result(True)

        self._event_helper.fire_event(
            ConnectionListener.on_connection_closed, self,
            self._state == ConnectorState.WAITING_TO_RECONNECT)

    def _reconnect(self):
        """Will be invoked by the IOLoop timer if the connection is
        closed. See the on_connection_closed method.

        """
        if not self._stopping:
            # Create a new connection
            self.connect()



from past.builtins import basestring
import re
import typing

__all__ = ['BroadcastFilter']


class BroadcastFilter(object):
    def __init__(self, subscriber, subject=None, sender=None):
        self._subscriber = subscriber
        self._subject_filters = []
        self._sender_filters = []
        if subject is not None:
            self.add_subject_filter(subject)
        if sender is not None:
            self.add_sender_filter(sender)

    def __call__(self, body, sender=None, subject=None, correlation_id=None):
        if subject is not None and self._subject_filters and \
                not any([check(subject) for check in self._subject_filters]):
            return

        if sender is not None and self._sender_filters and \
                not any([check(sender) for check in self._sender_filters]):
            return

        self._subscriber(body, sender, subject, correlation_id)

    def add_subject_filter(self, subject_filter):
        self._subject_filters.append(self._ensure_filter(subject_filter))

    def add_sender_filter(self, sender_filter):
        self._sender_filters.append(self._ensure_filter(sender_filter))

    def _ensure_filter(self, filter_value):
        if isinstance(filter_value, basestring):
            return re.compile(filter_value.replace('.', '\.').replace('*', '.*')).match
        elif isinstance(filter_value, typing.re.Pattern):
            return filter_value.match
        else:
            return lambda val: val == filter_value

    def _make_regex(self, filter_str):
        """
        :param filter_str: The filter string
        :type filter_str: basestring
        :return: The regular expression object
        """
        return re.compile(filter_str.replace('.', '\.'))



# -*- coding: utf-8 -*-

from setuptools import setup

__author__ = "Martin Uhrin"
__license__ = "GPLv3 and MIT, see LICENSE file"
__contributors__ = "Sebastiaan Huber"

about = {}
with open('kiwipy/version.py') as f:
    exec (f.read(), about)

setup(
    name="kiwipy",
    version=about['__version__'],
    description='A python remote communications library',
    long_description=open('README.md').read(),
    url='https://github.com/muhrin/kiwipy.git',
    author='Martin Uhrin',
    author_email='martin.uhrin@gmail.com',
    license=__license__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='communication messaging rpc broadcast',
    # Abstract dependencies.  Concrete versions are listed in
    # requirements.txt
    # See https://caremad.io/2013/07/setup-vs-requirement/ for an explanation
    # of the difference and
    # http://blog.miguelgrinberg.com/post/the-package-dependency-blues
    # for a useful dicussion
    install_requires=[
        'tornado',
        'future'
    ],
    extras_require={
        'rmq': ['pika', 'tornado', 'pyyaml'],
        'dev': ['pytest', 'pytest-cov', 'ipython', 'twine'],
        ':python_version<"3.5"': ['typing'],
        ':python_version<"3.4"': ['enum34'],
        ':python_version<"3.2"': ['backports.tempfile']
    },
    packages=['kiwipy', 'kiwipy.rmq'],
    test_suite='test'
)



from __future__ import absolute_import
import logging

from .communications import *
from .filters import *
from .futures import *
from .local import *
from .utils import *
from .version import *

__all__ = (futures.__all__ + communications.__all__ + version.__all__ + utils.__all__ +
           local.__all__ + filters.__all__)


# Do this se we don't get the "No handlers could be found..." warnings that will be produced
# if a user of this library doesn't set any handlers. See
# https://docs.python.org/3.1/library/logging.html#library-config
# for more details
class NullHandler(logging.Handler):
    def emit(self, record):
        pass


logging.getLogger("kiwipy").addHandler(NullHandler())



import unittest
import uuid

import kiwipy

from . import utils
from ..utils import CommunicatorTester

try:
    import pika
    from kiwipy import rmq
except ImportError:
    pika = None


@unittest.skipIf(not pika, "Requires pika library and RabbitMQ")
class TestCommunicator(CommunicatorTester, utils.TestCaseWithLoop):
    def create_communicator(self):
        self.connector = rmq.RmqConnector('amqp://guest:guest@localhost:5672/', loop=self.loop)
        self.exchange = "{}.{}".format(self.__class__.__name__, uuid.uuid4())
        self.task_queue = "{}.{}".format(self.__class__.__name__, uuid.uuid4())

        communicator = rmq.RmqCommunicator(
            self.connector,
            exchange_name=self.exchange,
            task_queue=self.task_queue,
            testing_mode=True
        )

        communicator.connect()
        return communicator

    def destroy_communicator(self, communicator):
        communicator.disconnect()


@unittest.skipIf(not pika, "Requires pika library and RabbitMQ")
class TestCommunicatorDroppyConnection(utils.TestCaseWithLoop):
    def setUp(self):
        super(TestCommunicatorDroppyConnection, self).setUp()

        self.connector = rmq.RmqConnector('amqp://guest:guest@localhost:5672/', loop=self.loop)
        self.exchange = "{}.{}".format(self.__class__.__name__, uuid.uuid4())
        self.task_queue = "{}.{}".format(self.__class__.__name__, uuid.uuid4())

        self._communicator = rmq.RmqCommunicator(
            self.connector,
            exchange_name=self.exchange,
            task_queue=self.task_queue,
            testing_mode=True
        )

        self._communicator.connect()

    def tearDown(self):
        self._communicator.disconnect()
        self.connector.disconnect()

        # Have to call super after because this closes the loop
        super(TestCommunicatorDroppyConnection, self).tearDown()

    def test_connect_disconnect(self):
        self.loop.run_sync(lambda: self.connector.connect())



import unittest
import uuid

import kiwipy
from . import utils

try:
    import pika
    from kiwipy import rmq
except ImportError:
    pika = None


@unittest.skipIf(not pika, "Requires pika library and RabbitMQ")
class TestTaskControllerAndRunner(utils.TestCaseWithLoop):
    def setUp(self):
        super(TestTaskControllerAndRunner, self).setUp()

        self.connector = rmq.RmqConnector('amqp://guest:guest@localhost:5672/', loop=self.loop)
        self.exchange_name = "{}.{}".format(self.__class__.__name__, uuid.uuid4())
        self.queue_name = "{}.{}.tasks".format(self.__class__.__name__, uuid.uuid4())

        self.subscriber = rmq.RmqTaskSubscriber(
            self.connector,
            exchange_name=self.exchange_name,
            task_queue_name=self.queue_name,
            testing_mode=True)
        self.publisher = rmq.RmqTaskPublisher(
            self.connector,
            exchange_name=self.exchange_name,
            task_queue_name=self.queue_name,
            testing_mode=True)

        self.loop.run_sync(lambda: self.subscriber.connect())
        self.loop.run_sync(lambda: self.publisher.connect())

    def tearDown(self):
        # Close the connector before calling super because it will
        # close the loop
        self.connector.close()
        super(TestTaskControllerAndRunner, self).tearDown()

    # TODO: Test publisher/subscriber
